if [ -z ${CONDA_BUILD+x} ]; then
	source /opt/concourse/worker/volumes/live/a8f76bc5-c7da-49cf-4117-4ff903c26135/volume/pymongo_1559921888537/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt